#include "io.h"
#include "term.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <termios.h>

int xgetc(int i);
void xputc(char c);
void writetotape(char c);
void writetoprinter(char c);
int readfromtape(void);

byte altair = 0;
byte ctrlz = 0;

byte mem[65536];
byte disk[4][77][26][128];

byte io_drive = 0;
byte io_track = 0;
byte io_sector = 0;
word io_dma;
byte io_diskstatus;

FILE *tapeoutfile = NULL;
FILE *tapeinfile = NULL;
FILE *printoutfile = NULL;

char bdos_prog[62] = {
	0x76, 0x00, 0x00, 0x00, 0x00, 0x32, 0x01, 0x00,
	0x78, 0x32, 0x02, 0x00, 0xeb, 0x22, 0x03, 0x00,
	0xeb, 0x79, 0xfe, 0x02, 0xca, 0x1d, 0x00, 0xfe,
	0x09, 0xca, 0x25, 0x00, 0x76, 0x3a, 0x01, 0x00,
	0xd3, 0x01, 0xc3, 0x31, 0x00, 0x1a, 0xfe, 0x24,
	0xca, 0x31, 0x00, 0xd3, 0x01, 0x13, 0xc3, 0x25,
	0x00, 0x3a, 0x02, 0x00, 0x47, 0x3a, 0x01, 0x00,
	0xeb, 0x2a, 0x03, 0x00, 0xeb, 0xc9
};

void setaltair(void) {
	altair = 1;
}

void bdos_load(void) {
	int i;

	for(i = 0; i < 62; i++) {
		write_mem(i, bdos_prog[i]);
	}
}

int read_mem(word addr) {
	return mem[addr];
}

int write_mem(word address, byte value) {
	mem[address] = value;
	return 0;
}

int input(byte port) {
	static int c;
	static int keyfetched = 0;

	switch(port) {
		case 0:
			if(keyfetched == 1) {
				c = xgetc(0);
				if(c == 27) {
					return 1000;
				}
				if(c == EOF) {
					return(altair ? 0x01 : 0x00);
				} else {
					keyfetched = 0;
					return(altair ? 0x00 : 0xff);
				}
			} else {
				return(altair ? 0x00 : 0xff);
			}
		case 1:
			if(keyfetched == 1)
				c = EOF;
			if(c == EOF)
				c = xgetc(1);
			if(c == 27) {
				return 1000;
			}
			keyfetched = 1;
			return(c == '\n' ? '\r' : c);
		case 4:
			return 0xff;
		case 5:
			return readfromtape();
		case 6:
			return 0x00;
		case 7:
			return readfromtape();
		case 13:
			return 1;
		case 14:
			return io_diskstatus;
		case 255:
			return 0;
		default:
			printf("IN: port %d\n", port);
			return 0;
	}
} 

int output(byte port, byte value) {
	int i;	

	switch(port) {
		case 1:
			xputc(value);
			return 0;
	    case 3:
			writetoprinter(value);
			return 0;
		case 5:
			writetotape(value);
			return 0;
		case 7:
			writetotape(value);
			return 0;
		case 10: // 0x0a FDC drive
			io_drive = value;
			break;
		case 11: // 0x0b FDC track
			io_track = value;
			break;
		case 12: // 0x0c FDC sector
			io_sector = value;
			break;
		case 13: // 0x0d FDC command
			if(io_drive > 3) { /* Num of disks */
				io_diskstatus = 2; // illegal drive
			} else if(io_track >= 77) { /* Num of tracks */
				io_diskstatus = 2; // illegal track
			} else if(io_sector == 0 || io_sector > 26) {
				io_diskstatus = 3; // illegal sector
			} else if(value == 0) {        // read
				if(io_dma >= 65536 - 128) {
					io_diskstatus = 5;  // read error
				} else {
					for(i = 0; i < 128; i++) {
						mem[io_dma + i] =
							disk[io_drive][io_track][io_sector-1][i];
					}
				}
			} else if(value == 1) { // write
				if(io_dma >= 65536 - 128) {
					io_diskstatus = 6;  // read error
				} else {
					for(i = 0; i < 128; i++) {
						disk[io_drive][io_track][io_sector-1][i] =
							mem[io_dma + i];
					}
				}
			} else {
				io_diskstatus = 7;    // illegal command
			}
			break;
		case 15: // 0x0f DMA low
			io_dma = (io_dma & 0xff00) | value;
			break;
		case 16: // 0x10 DMA high
			io_dma = (io_dma & 0x00ff) | ((word) value << 8);
			break;
		default:
			printf("OUT: port %d %d\n", port, value);
	}
	return 0;
}

void resetINTE(void) {
	return;
}

void setINTE(void) {
	return;
}

int readdisk(char *filename, int drive) {
	FILE *fp;
	int i, j, k;
	int c;

	if(!(fp = fopen(filename, "r"))) {
		fprintf(stderr, "Error: non-existing file - %s\n", filename);
		exit(2);
	}
	for(i=0; i<77; i++) {
		for(j=0; j<26; j++) {
			for(k = 0; k < 128; k++) {
				c = fgetc(fp);
				if(feof(fp)) {
					fprintf(stderr, "Error: Diskfile too short - %s\n",
						filename);
					break;
				}
				disk[drive][i][j][k] = (byte) c;
			}
		}
	}
	fclose(fp);
	return 0;
}

int writedisk(char *filename, int drive) {
	FILE *fp;
	int i, j, k;

	if(!(fp = fopen(filename, "w"))) {
		fprintf(stderr, "Error: writing error - %s\n", filename);
		exit(2);
	}
	for(i=0; i<77; i++) {
		for(j=0; j<26; j++) {
			for(k = 0; k < 128; k++) {
				fputc((int) disk[drive][i][j][k], fp);
			}
		}
	}
	fclose(fp);
	return 0;
}

int bootload(int drive) {
	int i;

	for(i = 0; i < 128; i++) {
		mem[i] = disk[0][0][0][i];
	}
	return 0;
}

int readrom(char *filename, int start) {
	FILE *fp;
	int i;
	int c;

	if(!(fp = fopen(filename, "r"))) {
		fprintf(stderr, "Error: non-existing file - %s\n", filename);
		exit(2);
	}
	for(i = 0; i < 65535 - start; i++) {
		c = getc(fp);
		if(feof(fp)) {
			break;
		}
		mem[i + start] = (byte) c;
	}
	fclose(fp);
	return 0;
}

int xgetc(int i) {
	char c;
	fflush(stdout);
	if(i == 1) {
		setcterm(1);
		if(read(STDIN_FILENO, &c, 1))
			;
		setcterm(0);
	}
	if(i == 0 && read(STDIN_FILENO, &c, 1) <= 0) {
		return EOF;
	} else if(c == 127) {
		if(altair) {
			return '_';
		} else {
			return 8;
		}
	} else if(c == 8) {
		return 127;
	} else if(c == 27) {
		ctrlz = 1;
		return 27;
	} else if(altair && c >= 97 && c <= 122) {
		return c - 32;
	} else {
		return c;
	}
}

void xputc(char c) {
	if(altair) {
		if(c == '_') {
			printf("\b \b");
			fflush(stdout);
			return;
		}
	}
	if(c < 0) {
		c = c + 128;
	}
	if(write(STDOUT_FILENO, &c, 1))
		;
}

void writetotape(char c) {
	if(!tapeoutfile) {
		if((tapeoutfile = fopen("output.aux","w")) == NULL) {
			return;
		}
	}
	if(fputc(c, tapeoutfile))
		;
}

void writetoprinter(char c) {
	if(!printoutfile) {
		if((printoutfile = fopen("output.lst","w")) == NULL) {
			return;
		}
	}
	if(fputc(c, printoutfile))
		;
}

int readfromtape(void) {
	int c;
	if(!tapeinfile) {
		if((tapeinfile = fopen("input.aux","r")) == NULL) {
			return 1000;
		}
	}
	c = fgetc(tapeinfile);
	if(c==EOF) {
		return 1000;
	}
	return c;
}

int getctrlz (void) {
	if (ctrlz) {
		ctrlz = 0;
		return 1;
	}
	return 0;
}

