#include "term.h"
#include "io.h"
#include "mainloop.h"
#include "8080.h"

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>

int main(int argc, char **argv) {
	char disk[4][80];
	int start = 0;
	int bdos = 0;
	float speed = 0;
	char rom[80];
	int c;
	int i;
	int quit_signal;

	for(i = 0; i < 4; i++) {
		disk[i][0] = 0;
	}
	rom[0] = 0;

	while((c = getopt(argc, argv, "a:b:c:d:f:s:tz")) != -1) {
		switch(c) {
			case 'a':
				memcpy(disk[0], optarg,
					80 * sizeof(char));
				break;
			case 'b':
				memcpy(disk[1], optarg,
					80 * sizeof(char));
				break;
			case 'c':
				memcpy(disk[2], optarg,
					80 * sizeof(char));
				break;
			case 'd':
				memcpy(disk[3], optarg,
					80 * sizeof(char));
				break;
			case 'f':
				if(!sscanf(optarg, "%f", &speed)) {
					fprintf(stderr,
						"No valid speed: %s\n",
						optarg);
					exit(1);
				}
				break;
			case 's':
				if(!sscanf(optarg, "%x", &start)) {
					fprintf(stderr,
						"No valid address: %s\n",
						optarg);
					exit(1);
				}
				break;
			case 't':
				setaltair();
				break;
			case 'z':
				bdos = 1;
				break;
			case '?':
				if(optopt == 'a' || optopt == 'b'
					|| optopt == 'c'
					|| optopt == 'd'
					|| optopt == 's') {
					fprintf(stderr,
						"Option -%c requires an argument.\n", optopt);
				} else if(isprint (optopt)) {
					fprintf(stderr, "Unknown option `-%c'.\n", optopt);
				} else {
					fprintf(stderr,
						"Unknown option character `\\x%x'.\n", optopt);
				}
				exit(1);
			default:
				abort();
		}
	}

	if(optind < argc) {
		memcpy(rom, argv[optind], 80 * sizeof(char));
	}

	for(i = 0; i < 4; i++) {
		if(disk[i][0]) {
			readdisk(disk[i], i);
		}
	}

	if(start > 65535) {
		fprintf(stderr, "Out of bounds.\n");
		exit(1);
	}

	if(rom[0]) {
		readrom(rom, start);
		if(bdos) {
			bdos_load();
		}
	} else if(disk[0]) {
		bootload(0);
	} else {
		fprintf(stderr, "Nothing to load from...\n");
		exit(1);
	}

	setterm();
	quit_signal = mainloop(start, speed);
	resetterm();

	for(i = 0; i < 4; i++) {
		if(disk[i][0]) {
			writedisk(disk[i], i);
		}
	}

	printf("\n");
	if(quit_signal != HALTED) {
		return quit_signal;
	}
	return 0;
}

